@@ -2,7 +2,6 @@
 # -*- coding: latin-1 -*-
 # ****************************************************************************
 # * Software: FPDF for python                                                *
-# * Version:  2.0.6                                                          *
 # * Date:     2010-09-10                                                     *
 # * Last update: 2020-10-25                                                  *
 # * License:  LGPL v3.0                                                      *
@@ -20,11 +19,13 @@ updated here (above and below in variable).
 """
 from __future__ import division, with_statement
 
+from contextlib import contextmanager
 from datetime import datetime
 from collections import OrderedDict as o_dict
 from functools import wraps
 from future.utils import raise_from
 import errno
+import logging
 import math
 import os
 import re
@@ -57,8 +58,10 @@ from .util.syntax import (
     create_stream as pdf_stream
 )
 
+LOGGER = logging.getLogger(__name__)
+
 # Global variables
-FPDF_VERSION = '2.0.1'
+FPDF_VERSION = '2.1.0'
 FPDF_FONT_DIR = os.path.join(os.path.dirname(__file__), 'font')
 FPDF_CACHE_MODE = 0  # 0 - in same folder, 1 - none, 2 - hash
 FPDF_CACHE_DIR = None
@@ -131,7 +134,7 @@ class FPDF(object):
         self.offsets = {}               # array of object offsets
         self.page = 0                   # current page number
         self.n = 2                      # current object number
-        self.buffer = ''                # buffer holding in-memory PDF
+        self.buffer = bytearray()       # buffer holding in-memory PDF
         self.pages = {}                 # array containing pages and metadata
         self.state = 0                  # current document state
         self.fonts = {}                 # array of used fonts
@@ -751,13 +754,15 @@ class FPDF(object):
 
         self.links[link] = [page, y]
 
-    def link(self, x, y, w, h, link):
+    def link(self, x, y, w, h, link, alt_text=''):
         "Put a link on the page"
         if self.page not in self.page_links:
             self.page_links[self.page] = []
         self.page_links[self.page] += [(
             x * self.k, self.h_pt - y * self.k,
-            w * self.k, h             * self.k, link
+            w * self.k, h             * self.k,
+            link,
+            alt_text
         )]
 
     @check_page
@@ -765,7 +770,7 @@ class FPDF(object):
         "Output a string"
         txt = self.normalize_text(txt)
         if (self.unifontsubset):
-            txt2 = escape_parens(UTF8ToUTF16BE(txt, False))
+            txt2 = UTF8ToUTF16BE(escape_parens(txt), False)
             for uni in UTF8StringToArray(txt):
                 self.current_font['subset'].append(uni)
         else:
@@ -782,6 +787,7 @@ class FPDF(object):
 
     @check_page
     def rotate(self, angle, x = None, y = None):
+        warnings.warn('rotate() can produces malformed PDFs and is deprecated. Use the rotation() context manager instead.', PendingDeprecationWarning)
         if x is None: x = self.x
         if y is None: y = self.y
 
@@ -799,6 +805,18 @@ class FPDF(object):
                         c, s, -s, c, cx, cy, -cx, -cy)
             self._out(s)
 
+    @check_page
+    @contextmanager
+    def rotation(self, angle, x=None, y=None):
+        if x is None: x = self.x
+        if y is None: y = self.y;
+        angle *= math.pi / 180
+        c, s = math.cos(angle), math.sin(angle)
+        cx, cy = x*self.k, (self.h-y)*self.k
+        self._out(sprintf('q %.5F %.5F %.5F %.5F %.2F %.2F cm 1 0 0 1 %.2F %.2F cm\n', c, s, -s, c, cx, cy, -cx, -cy))
+        yield
+        self._out('Q\n')
+
     @property
     def accept_page_break(self):
         "Accept automatic page break or not"
@@ -1159,7 +1177,8 @@ class FPDF(object):
                       h, substr(s, j), 0, 0, '', 0, link)
 
     @check_page
-    def image(self, name, x = None, y = None, w = 0, h = 0, type = '', link = ''):  # noqa: E501
+    def image(self, name, x=None, y=None, w=0,h=0,type='',link='', is_mask=False, mask_image=None):
+    #def image(self, name, x = None, y = None, w = 0, h = 0, type = '', link = ''):  # noqa: E501
         "Put an image on the page"
         if name not in self.images:
             info = get_img_info(image_parsing_load_resource(name))
@@ -1198,6 +1217,8 @@ class FPDF(object):
         if (link):
             self.link(x, y, w, h, link)
 
+        return info
+
     @check_page
     def ln(self, h = None):
         "Line Feed; default value is last cell height"
@@ -1243,24 +1264,17 @@ class FPDF(object):
         if (dest == ''):
             if (name == ''): dest = 'I'
             else:            dest = 'F'
-
-        if PY3K:
-            # manage binary data as latin1 until PEP461 or similar is
-            # implemented
-            buffer = self.buffer.encode("latin1")
-        else:
-            buffer = self.buffer
         if dest in ('I', 'D'):
             # Python < 3 writes byte data transparently without "buffer"
             stdout = getattr(sys.stdout, 'buffer', sys.stdout)
-            stdout.write(buffer)
+            stdout.write(self.buffer)
         elif dest == 'F':
             # Save to local file
             with open(name, 'wb') as f:
-                f.write(buffer)
+                f.write(self.buffer)
         # Return as a byte string
-        elif dest == 'S': return buffer
-
+        elif dest == 'S':
+            return self.buffer
         else:
             fpdf_error('Incorrect output destination: ' + dest)
 
@@ -1292,9 +1306,8 @@ class FPDF(object):
                     self.pages[n]["content"].replace(alias, r)
             # Now repeat for no pages in non-subset fonts
             for n in range(1, nb + 1):
-                self.pages[n]["content"] = \
-                    self.pages[n]["content"] \
-                        .replace(self.str_alias_nb_pages, str(nb))
+                self.pages[n]["content"].replace(self.str_alias_nb_pages.encode(),
+                        str(nb).encode())
         if self.def_orientation == 'P':
             dw_pt = self.dw_pt
             dh_pt = self.dh_pt
@@ -1339,12 +1352,12 @@ class FPDF(object):
 
                     # Dest type ending of annotation entry
                     else:
+                        assert pl[4] in self.links, 'Page {} has a link with an invalid index: {} (doc #links={})'.format(n, pl[4], len(self.links))
                         l = self.links[pl[4]]
                         # if l[0] in self.orientation_changes: h = w_pt
                         # else:                                h = h_pt
-                        h = h_pt
                         annots += sprintf('/Dest [%d 0 R /XYZ 0 %.2f null]>>',
-                                          1 + 2 * l[0], h - l[1] * self.k)
+                                          1 + 2 * l[0], h_pt - l[1] * self.k)
                 
                 # End links list
                 self._out(annots + ']')
@@ -1357,10 +1370,7 @@ class FPDF(object):
             # Page content
             content = self.pages[n]["content"]
             if self.compress:
-                # manage binary data as latin1 until PEP461 or similar is
-                # implemented
-                p = content.encode("latin1") if PY3K else content
-                p = zlib.compress(p)
+                p = zlib.compress(content)
             else:
                 p = content
             self._newobj()
@@ -1798,16 +1808,19 @@ class FPDF(object):
         self._out('>>')
 
     def _putresources(self):
-        self._putfonts()
-        self._putimages()
+        with self._trace_size('resources.fonts'):
+            self._putfonts()
+        with self._trace_size('resources.images'):
+            self._putimages()
 
         # Resource dictionary
-        self.offsets[2] = len(self.buffer)
-        self._out('2 0 obj')
-        self._out('<<')
-        self._putresourcedict()
-        self._out('>>')
-        self._out('endobj')
+        with self._trace_size('resources.dict'):
+            self.offsets[2] = len(self.buffer)
+            self._out('2 0 obj')
+            self._out('<<')
+            self._putresourcedict()
+            self._out('>>')
+            self._out('endobj')
 
     def _putinfo(self):
         ts = lambda a: enclose_in_parens(a)
@@ -1831,8 +1844,7 @@ class FPDF(object):
             date_string = datetime.now().strftime('%Y%m%d%H%M%S')
         info_d[pdf_name('CreationDate')] = ts('D:' + date_string)
 
-        return pdf_d(info_d, open_dict='<<\n', close_dict='\n>>',
-            has_empty_fields=True)
+        self._out(pdf_d(info_d, open_dict='', close_dict='', has_empty_fields=True))
 
     def _putcatalog(self):
         catalog_d = o_dict()
@@ -1863,7 +1875,7 @@ class FPDF(object):
         if self.layout_mode in layout_names:
             catalog_d[pdf_name('PageLayout')] = layout_names[self.layout_mode]
 
-        return pdf_d(catalog_d, open_dict='<<\n', close_dict='\n>>')
+        self._out(pdf_d(catalog_d, open_dict='', close_dict=''))
 
     def _putheader(self):
         self._out('%PDF-' + self.pdf_version)
@@ -1874,37 +1886,48 @@ class FPDF(object):
         self._out('/Info ' + pdf_ref(self.n - 1))
 
     def _enddoc(self):
-        self._putheader()
-        self._putpages()
-        self._putresources()
-        # Info
-        self._newobj()
-        self._out(self._putinfo())
-        self._out('endobj')
-        # Catalog
-        self._newobj()
-        self._out(self._putcatalog())
-        self._out('endobj')
-        # Cross-ref
-        o = len(self.buffer)
-        self._out('xref')
-        self._out('0 ' + (str(self.n + 1)))
-        self._out('0000000000 65535 f ')
-        for i in range(1, self.n + 1):
-            self._out(sprintf('%010d 00000 n ', self.offsets[i]))
-        # Trailer
-        self._out('trailer')
-        self._out('<<')
-        self._puttrailer()
-        self._out('>>')
-        self._out('startxref')
-        self._out(o)
+        LOGGER.debug('Final doc sections size summary:')
+        with self._trace_size('header'):
+            self._putheader()
+        with self._trace_size('pages'):
+            self._putpages()
+        self._putresources()  # trace_size is performed inside
+        #Info
+        with self._trace_size('info'):
+            self._newobj()
+            self._out('<<')
+            self._putinfo()
+            self._out('>>')
+            self._out('endobj')
+        #Catalog
+        with self._trace_size('catalog'):
+            self._newobj()
+            self._out('<<')
+            self._putcatalog()
+            self._out('>>')
+            self._out('endobj')
+        #Cross-ref
+        with self._trace_size('xref'):
+            o=len(self.buffer)
+            self._out('xref')
+            self._out('0 '+(str(self.n+1)))
+            self._out('0000000000 65535 f ')
+            for i in range(1,self.n+1):
+                self._out(sprintf('%010d 00000 n ',self.offsets[i]))
+        #Trailer
+        with self._trace_size('trailer'):
+            self._out('trailer')
+            self._out('<<')
+            self._puttrailer()
+            self._out('>>')
+            self._out('startxref')
+            self._out(o)
         self._out('%%EOF')
         self.state = 3
 
     def _beginpage(self, orientation, format, same):
         self.page             += 1
-        self.pages[self.page]  = {"content": ""}
+        self.pages[self.page]  = {"content": bytearray()}
         self.state             = 2
         self.x                 = self.l_margin
         self.y                 = self.t_margin
@@ -1973,10 +1996,114 @@ class FPDF(object):
         elif not isinstance(s, basestring):
             s = str(s)
         if (self.state == 2):
-            self.pages[self.page]["content"] += (s + "\n")
+            #self.pages[self.page]["content"] += (s + "\n")
+            self.pages[self.page]["content"] += (s.encode("latin1") + b"\n")
         else:
-            self.buffer += (s + "\n")
+            self.buffer += (s.encode("latin1") + b"\n")
+
+    @check_page
+    def interleaved2of5(self, txt, x, y, w=1.0, h=10.0):
+        "Barcode I2of5 (numeric), adds a 0 if odd lenght"
+        narrow = w / 3.0
+        wide = w
+
+        # wide/narrow codes for the digits
+        bar_char={'0': 'nnwwn', '1': 'wnnnw', '2': 'nwnnw', '3': 'wwnnn',
+                  '4': 'nnwnw', '5': 'wnwnn', '6': 'nwwnn', '7': 'nnnww',
+                  '8': 'wnnwn', '9': 'nwnwn', 'A': 'nn', 'Z': 'wn'}
+
+        self.set_fill_color(0)
+        code = txt
+        # add leading zero if code-length is odd
+        if len(code) % 2 != 0:
+            code = '0' + code
+
+        # add start and stop codes
+        code = 'AA' + code.lower() + 'ZA'
+
+        for i in range(0, len(code), 2):
+            # choose next pair of digits
+            char_bar = code[i]
+            char_space = code[i+1]
+            # check whether it is a valid digit
+            if not char_bar in bar_char.keys():
+                raise RuntimeError ('Char "%s" invalid for I25: ' % char_bar)
+            if not char_space in bar_char.keys():
+                raise RuntimeError ('Char "%s" invalid for I25: ' % char_space)
+
+            # create a wide/narrow-seq (first digit=bars, second digit=spaces)
+            seq = ''
+            for s in range(0, len(bar_char[char_bar])):
+                seq += bar_char[char_bar][s] + bar_char[char_space][s]
+
+            for bar in range(0, len(seq)):
+                # set line_width depending on value
+                if seq[bar] == 'n':
+                    line_width = narrow
+                else:
+                    line_width = wide
+
+                # draw every second value, the other is represented by space
+                if bar % 2 == 0:
+                    self.rect(x, y, line_width, h, 'F')
+
+                x += line_width
+
+
+    @check_page
+    def code39(self, txt, x, y, w=1.5, h=5.0):
+        """Barcode 3of9"""
+        dim = {'w': w, 'n': w/3.}
+        chars = {
+            '0': 'nnnwwnwnn', '1': 'wnnwnnnnw', '2': 'nnwwnnnnw',
+            '3': 'wnwwnnnnn', '4': 'nnnwwnnnw', '5': 'wnnwwnnnn',
+            '6': 'nnwwwnnnn', '7': 'nnnwnnwnw', '8': 'wnnwnnwnn',
+            '9': 'nnwwnnwnn', 'A': 'wnnnnwnnw', 'B': 'nnwnnwnnw',
+            'C': 'wnwnnwnnn', 'D': 'nnnnwwnnw', 'E': 'wnnnwwnnn',
+            'F': 'nnwnwwnnn', 'G': 'nnnnnwwnw', 'H': 'wnnnnwwnn',
+            'I': 'nnwnnwwnn', 'J': 'nnnnwwwnn', 'K': 'wnnnnnnww',
+            'L': 'nnwnnnnww', 'M': 'wnwnnnnwn', 'N': 'nnnnwnnww',
+            'O': 'wnnnwnnwn', 'P': 'nnwnwnnwn', 'Q': 'nnnnnnwww',
+            'R': 'wnnnnnwwn', 'S': 'nnwnnnwwn', 'T': 'nnnnwnwwn',
+            'U': 'wwnnnnnnw', 'V': 'nwwnnnnnw', 'W': 'wwwnnnnnn',
+            'X': 'nwnnwnnnw', 'Y': 'wwnnwnnnn', 'Z': 'nwwnwnnnn',
+            '-': 'nwnnnnwnw', '.': 'wwnnnnwnn', ' ': 'nwwnnnwnn',
+            '*': 'nwnnwnwnn', '$': 'nwnwnwnnn', '/': 'nwnwnnnwn',
+            '+': 'nwnnnwnwn', '%': 'nnnwnwnwn',
+        }
+        self.set_fill_color(0)
+        for c in txt.upper():
+            if c not in chars:
+                raise RuntimeError('Invalid char "%s" for Code39' % c)
+            for i, d in enumerate(chars[c]):
+                if i % 2 == 0:
+                    self.rect(x, y, dim[d], h, 'F')
+                x += dim[d]
+            x += dim['n']
+
+    @check_page
+    @contextmanager
+    def rect_clip(self, x, y, w, h):
+        self._out(sprintf('q %.2f %.2f %.2f %.2f re W n\n',x*self.k,(self.h-(y+h))*self.k,w*self.k,h*self.k))
+        yield
+        self._out('Q\n')
+
+    @contextmanager
+    def _trace_size(self, label):
+        prev_size = len(self.buffer)
+        yield
+        LOGGER.debug('- %s.size: %s', label, _sizeof_fmt(len(self.buffer) - prev_size))
+
+
+def _sizeof_fmt(num, suffix='B'):
+    # Recipe from: https://stackoverflow.com/a/1094933/636849
+    for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:
+        if abs(num) < 1024.0:
+            return "%3.1f%s%s" % (num, unit, suffix)
+        num /= 1024.0
+    return "%.1f%s%s" % (num, 'Yi', suffix)
 
 __all__ = [
     'FPDF', 'load_cache', 'get_page_format', 'PAGE_FORMATS'
 ]
+
